@@ -41,16 +41,17 @@ logger = logging.getLogger("pykeepass")
 
 
 class KDBX3Tests(unittest.TestCase):
-    database = 'test3.kdbx'
+    database = os.path.join(base_dir, 'test3.kdbx')
     password = 'password'
-    keyfile = 'test3.key'
+    keyfile = os.path.join(base_dir, 'test3.key')
+
+    database_tmp = os.path.join(base_dir, 'test3_tmp.kdbx')
+    keyfile_tmp = os.path.join(base_dir, 'test3_tmp.key')
 
     # get some things ready before testing
     def setUp(self):
-        shutil.copy(
-            os.path.join(base_dir, self.database),
-            os.path.join(base_dir, 'change_creds.kdbx')
-        )
+        shutil.copy(self.database, self.database_tmp)
+        shutil.copy(self.keyfile, self.keyfile_tmp)
         self.kp = PyKeePass(
             os.path.join(base_dir, self.database),
             password=self.password,
@@ -58,16 +59,23 @@ class KDBX3Tests(unittest.TestCase):
         )
         # for tests which modify the database, use this
         self.kp_tmp = PyKeePass(
-            os.path.join(base_dir, 'change_creds.kdbx'),
+            os.path.join(base_dir, self.database_tmp),
             password=self.password,
-            keyfile=os.path.join(base_dir, self.keyfile)
+            keyfile=os.path.join(base_dir, self.keyfile_tmp)
         )
 
+    def tearDown(self):
+        os.remove(self.keyfile_tmp)
+        os.remove(self.database_tmp)
+
 
 class KDBX4Tests(KDBX3Tests):
-    database = 'test4.kdbx'
+    database = os.path.join(base_dir, 'test4.kdbx')
     password = 'password'
-    keyfile = 'test4.key'
+    keyfile = os.path.join(base_dir, 'test4.key')
+
+    database_tmp = os.path.join(base_dir, 'test4_tmp.kdbx')
+    keyfile_tmp = os.path.join(base_dir, 'test4_tmp.key')
 
 
 class EntryFindTests3(KDBX3Tests):
@@ -143,6 +151,11 @@ class EntryFindTests3(KDBX3Tests):
         results = self.kp.find_entries(autotype_enabled=True)
         self.assertEqual(len(results), len(self.kp.entries) - 1)
 
+    def test_find_entries_by_otp(self):
+        results = self.kp.find_entries(otp='otpsecret', regex=True, flags='i')
+        self.assertEqual(len(results), 1)
+        self.assertEqual('foobar_entry', results[0].title)
+
     def test_find_entries(self):
         results = self.kp.find_entries(title='Root_entry', regex=True)
         self.assertEqual(len(results), 0)
@@ -230,6 +243,7 @@ class EntryFindTests3(KDBX3Tests):
         self.assertEqual(results.notes, unique_str + 'notes')
         self.assertEqual(results.tags, [unique_str + 'tags'])
         self.assertTrue(results.uuid != None)
+        self.assertTrue(results.autotype_sequence is None)
         # convert naive datetime to utc
         expiry_time_utc = expiry_time.replace(tzinfo=tz.gettz()).astimezone(tz.gettz('UTC'))
         self.assertEqual(results.icon, icons.KEY)
@@ -425,6 +439,7 @@ class EntryTests3(KDBX3Tests):
             url='url',
             notes='notes',
             tags='tags',
+            otp='otp',
             expires=True,
             expiry_time=time,
             icon=icons.KEY,
@@ -437,6 +452,7 @@ class EntryTests3(KDBX3Tests):
         self.assertEqual(entry.url, 'url')
         self.assertEqual(entry.notes, 'notes')
         self.assertEqual(entry.tags, ['tags'])
+        self.assertEqual(entry.otp, 'otp')
         self.assertEqual(entry.expires, True)
         self.assertEqual(entry.expiry_time,
                          time.replace(tzinfo=tz.gettz()).astimezone(tz.gettz('UTC')))
@@ -489,6 +505,7 @@ class EntryTests3(KDBX3Tests):
         entry.icon = icons.GLOBE
         entry.set_custom_property('foo', 'bar')
         entry.set_custom_property('multiline', 'hello\nworld')
+        entry.otp = "otpsecret"
 
         self.assertEqual(entry.title, changed_string + 'title')
         self.assertEqual(entry.username, changed_string + 'username')
@@ -498,6 +515,7 @@ class EntryTests3(KDBX3Tests):
         self.assertEqual(entry.icon, icons.GLOBE)
         self.assertEqual(entry.get_custom_property('foo'), 'bar')
         self.assertEqual(entry.get_custom_property('multiline'), 'hello\nworld')
+        self.assertEqual(entry.otp, 'otpsecret')
         self.assertIn('foo', entry.custom_properties)
         entry.delete_custom_property('foo')
         self.assertEqual(entry.get_custom_property('foo'), None)
@@ -638,7 +656,7 @@ class EntryHistoryTests3(KDBX3Tests):
             for item in hist:
                 self.assertTrue(item.is_a_history_entry)
                 self.assertEqual(item.group, entry.group)
-                self.assertTrue(str(item).startswith('[History of:'))
+                self.assertTrue(str(item).startswith('HistoryEntry:'))
 
         # here history items are expected
         res2 = self.kp.find_entries(title=prefix + 'title', history=True)
@@ -698,7 +716,6 @@ class EntryHistoryTests3(KDBX3Tests):
             for item in hist:
                 self.assertTrue(item.is_a_history_entry)
                 self.assertEqual(item.group, entry.group)
-                self.assertTrue(str(item).startswith('[History of:'))
 
         res2 = self.kp.find_entries(title=changed + 'title', history=True)
         self.assertEqual(len(res2), 6)
@@ -706,6 +723,16 @@ class EntryHistoryTests3(KDBX3Tests):
             if entry not in res1:
                 self.assertTrue(entry.is_a_history_entry)
 
+        # try deleting a history entry
+        h = e1.history[0]
+        self.assertIn(h, e1.history)
+        e1.delete_history(h)
+        self.assertNotIn(h, e1.history)
+
+        # delete all history
+        self.assertTrue(len(e1.history) > 0)
+        e1.delete_history(all=True)
+        self.assertTrue(len(e1.history) == 0)
 
 class GroupTests3(KDBX3Tests):
 
@@ -729,39 +756,26 @@ class GroupTests3(KDBX3Tests):
 
 class AttachmentTests3(KDBX3Tests):
     # get some things ready before testing
-    def setUp(self):
-        shutil.copy(
-            os.path.join(base_dir, self.database),
-            os.path.join(base_dir, 'test_attachment.kdbx')
-        )
-        self.open()
-
-    def open(self):
-        self.kp = PyKeePass(
-            os.path.join(base_dir, 'test_attachment.kdbx'),
-            password=self.password,
-            keyfile=os.path.join(base_dir, self.keyfile)
-        )
 
     def test_create_delete_binary(self):
         with self.assertRaises(BinaryError):
-            self.kp.delete_binary(999)
+            self.kp_tmp.delete_binary(999)
         with self.assertRaises(BinaryError):
-            e = self.kp.entries[0]
+            e = self.kp_tmp.entries[0]
             e.add_attachment(filename='foo.txt', id=123)
             e.attachments[0].binary
 
-        binary_id = self.kp.add_binary(b'Ronald McDonald Trump')
-        self.kp.save()
-        self.open()
-        self.assertEqual(self.kp.binaries[binary_id], b'Ronald McDonald Trump')
-        self.assertEqual(len(self.kp.attachments), 1)
+        binary_id = self.kp_tmp.add_binary(b'Ronald McDonald Trump')
+        self.kp_tmp.save()
+        self.kp_tmp.reload()
+        self.assertEqual(self.kp_tmp.binaries[binary_id], b'Ronald McDonald Trump')
+        self.assertEqual(len(self.kp_tmp.attachments), 1)
 
-        num_attach = len(self.kp.binaries)
-        self.kp.delete_binary(binary_id)
-        self.kp.save()
-        self.open()
-        self.assertEqual(len(self.kp.binaries), num_attach - 1)
+        num_attach = len(self.kp_tmp.binaries)
+        self.kp_tmp.delete_binary(binary_id)
+        self.kp_tmp.save()
+        self.kp_tmp.reload()
+        self.assertEqual(len(self.kp_tmp.binaries), num_attach - 1)
 
     def test_attachment_reference_decrement(self):
         e = self.kp.entries[0]
@@ -784,9 +798,6 @@ class AttachmentTests3(KDBX3Tests):
         self.assertEqual(a.id, binary_id)
         self.assertEqual(a.filename, 'test.txt')
 
-    def tearDown(self):
-        os.remove(os.path.join(base_dir, 'test_attachment.kdbx'))
-
 
 class PyKeePassTests3(KDBX3Tests):
 
@@ -809,9 +820,46 @@ class PyKeePassTests3(KDBX3Tests):
             first_line = f.readline()
             self.assertEqual(first_line, '<?xml version=\'1.0\' encoding=\'utf-8\' standalone=\'yes\'?>\n')
 
-    def tearDown(self):
-        os.remove(os.path.join(base_dir, 'change_creds.kdbx'))
+    def test_credchange(self):
+        """
+        - test rec/req boolean (expired, no expired, days=-1)
+        - test get/set days
+        - test cred set timer reset
+        """
 
+        required_days = 5
+        recommended_days = 5
+        unexpired_date = datetime.now() - timedelta(days=1)
+        expired_date = datetime.now() - timedelta(days=10)
+
+        self.kp.credchange_required_days = required_days
+        self.kp.credchange_recommended_days = recommended_days
+
+        # test not expired
+        self.kp.credchange_date = unexpired_date
+        self.assertFalse(self.kp.credchange_required)
+        self.assertFalse(self.kp.credchange_recommended)
+
+        # test expired
+        self.kp.credchange_date = expired_date
+        self.assertTrue(self.kp.credchange_required)
+        self.assertTrue(self.kp.credchange_recommended)
+
+        # test expiry disabled
+        self.kp.credchange_required_days = -1
+        self.kp.credchange_recommended_days = -1
+        self.assertFalse(self.kp.credchange_required)
+        self.assertFalse(self.kp.credchange_recommended)
+
+        # test credential update
+        self.kp.credchange_required_days = required_days
+        self.kp.credchange_recommended_days = recommended_days
+        self.kp.credchange_date = expired_date
+        self.assertTrue(self.kp.credchange_required)
+        self.assertTrue(self.kp.credchange_recommended)
+        self.kp.keyfile = 'foo'
+        self.assertFalse(self.kp.credchange_required)
+        self.assertFalse(self.kp.credchange_recommended)
 
 class BugRegressionTests3(KDBX3Tests):
     def test_issue129(self):
@@ -851,6 +899,43 @@ class BugRegressionTests3(KDBX3Tests):
         e = self.kp_tmp.find_entries(title='protect_test', first=True)
         self.assertEqual(e.password, 'pass')
 
+    def test_issue223(self):
+        # issue 223 - database is clobbered when kp.save() fails
+        # even if exception is caught
+
+        # test clobbering with file on disk
+        # change keyfile so database save fails
+        self.kp_tmp.keyfile = 'foo'
+        with self.assertRaises(Exception):
+            self.kp_tmp.save()
+        # try to open database
+        self.kp_tmp.keyfile = self.keyfile_tmp
+        PyKeePass(self.database_tmp, self.password, self.keyfile_tmp)
+        # reset test database
+        self.setUp()
+
+        # test clobbering with buffer
+        stream = BytesIO()
+        self.kp_tmp.save(stream)
+        stream.seek(0)
+
+        # change keyfile so database save fails
+        self.kp_tmp.keyfile = 'foo'
+        self.kp_tmp.password = ('invalid', 'type')
+        with self.assertRaises(Exception):
+            self.kp_tmp.save(stream)
+        stream.seek(0)
+        # try to open database
+        self.kp_tmp.keyfile = self.keyfile_tmp
+        PyKeePass(stream, self.password, self.keyfile_tmp)
+
+    def test_issue308(self):
+        # find_entries/find_groups() break when supplied None values directly
+
+        results = self.kp.find_entries(title='foobar_entry')
+        results2 = self.kp.find_entries(title='foobar_entry', username=None)
+
+        self.assertEqual(results, results2)
 
 
 class EntryFindTests4(KDBX4Tests, EntryFindTests3):
@@ -904,7 +989,10 @@ class KDBXTests(unittest.TestCase):
             os.path.join(base_dir, 'test3_transformed.kdbx'),     # KDBX v3 transformed_key open
             os.path.join(base_dir, 'test4_transformed.kdbx'),     # KDBX v4 transformed_key open
             stream,
-            os.path.join(base_dir, 'test4_aes_uncompressed.kdbx') # KDBX v4 AES uncompressed
+            os.path.join(base_dir, 'test4_aes_uncompressed.kdbx'),# KDBX v4 AES uncompressed
+            os.path.join(base_dir, 'test4_twofish_uncompressed.kdbx'),# KDBX v4 Twofish uncompressed
+            os.path.join(base_dir, 'test4_chacha20_uncompressed.kdbx'),# KDBX v4 ChaCha uncompressed
+            os.path.join(base_dir, 'test4_argon2id.kdbx'),        # KDBX v4 Argon2id
         ]
         filenames_out = [
             os.path.join(base_dir, 'test3.kdbx.out'),
@@ -918,8 +1006,9 @@ class KDBXTests(unittest.TestCase):
             os.path.join(base_dir, 'test4_transformed.kdbx.out'),
             BytesIO(),
             os.path.join(base_dir, 'test4_aes_uncompressed.kdbx.out'),
-            os.path.join(base_dir, 'test4_twofish_uncompressed.kdbx.out'),
-            os.path.join(base_dir, 'test4_chacha20_uncompressed.kdbx.out'),
+            os.path.join(base_dir, 'test4_twofish_uncompressed.kdbx.out'),# KDBX v4 Twofish uncompressed
+            os.path.join(base_dir, 'test4_chacha20_uncompressed.kdbx.out'),# KDBX v4 ChaCha uncompressed
+            os.path.join(base_dir, 'test4_argon2id.kdbx.out'),
         ]
         passwords = [
             'password',
@@ -935,6 +1024,7 @@ class KDBXTests(unittest.TestCase):
             'password',
             'password',
             'password',
+            'password',
         ]
         transformed_keys = [
             None,
@@ -950,6 +1040,7 @@ class KDBXTests(unittest.TestCase):
             None,
             None,
             None,
+            None,
         ]
         keyfiles = [
             'test3.key',
@@ -965,6 +1056,7 @@ class KDBXTests(unittest.TestCase):
             None,
             None,
             None,
+            None,
         ]
         encryption_algorithms = [
             'aes256',
@@ -980,6 +1072,7 @@ class KDBXTests(unittest.TestCase):
             'aes256',
             'twofish',
             'chacha20',
+            'aes256',
         ]
         kdf_algorithms = [
             'aeskdf',
@@ -995,6 +1088,7 @@ class KDBXTests(unittest.TestCase):
             'argon2',
             'argon2',
             'argon2',
+            'argon2id',
         ]
         versions = [
             (3, 1),
@@ -1010,6 +1104,7 @@ class KDBXTests(unittest.TestCase):
             (4, 0),
             (4, 0),
             (4, 0),
+            (4, 0),
         ]
 
         for (filename_in, filename_out, password, transformed_key,
@@ -1078,3 +1173,4 @@ class KDBXTests(unittest.TestCase):
 
 if __name__ == '__main__':
     unittest.main()
+
